import json

from google.auth.transport.requests import Request
from google.oauth2 import service_account
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import Resource, build

SCOPES = [
    "https://www.googleapis.com/auth/drive.metadata.readonly",
    "https://www.googleapis.com/auth/drive.readonly",
    "https://www.googleapis.com/auth/documents.readonly",
]


def load_creds():
    """
    Loads creds from a pickle file. If creds do not exist, requests for user authorization. Modified from the Google Drive API Python quickstart guide.

    Returns:
        creds (tuple): the creds for the google drive API
    """
    credentials = service_account.Credentials.from_service_account_file(
        "creds.json", scopes=SCOPES
    )
    return credentials



import gsimplify.auth
import tqdm
import os
import gsimplify.drive
import shutil
from googleapiclient.http import MediaIoBaseDownload
import gsimplify.templates
from typing import Union, List, Tuple
import git

from gsimplify.typedefs import FolderType


class Builder:
    def __init__(self, drive_str: str, template_dir: str, assets_dir: str = ""):
        self.drive_str = drive_str
        self.creds = gsimplify.auth.load_creds()
        self.drive = gsimplify.drive.Drive(drive_str, self.creds)
        self.templator = gsimplify.templates.Templates(template_dir)
        self.template_dir = template_dir
        self.assets_dir = assets_dir

        self.folders = self.drive.folders()
        self.file_tree = self.construct_folder_tree()
        self.navbar_items = self.construct_navbar()

    def adjust_relative_link(self, link: str):
        DRIVE_LINK_PREFIX = "https://docs.google.com/document/d/"
        DRIVE_LINK_END = "/"

        if DRIVE_LINK_PREFIX in link:
            doc_id = link.split(DRIVE_LINK_PREFIX)[1].split(DRIVE_LINK_END)[0]

            for doc in self.drive.docs():
                if doc_id == doc.id:
                    for folder in self.drive.folders():
                        if folder.id == doc.parents:
                            break
                    else:
                        print("folder not found!")
                        return link

                    path = f'{self.path_join(folder, "")}{doc.pointer}.html'

                    return path

        return link

    @staticmethod
    def path_join(folder: gsimplify.typedefs.FolderType, prefix: str = "./build/"):
        return prefix + "/".join(folder.path) + "/"

    def construct_folder_tree(
        self, path: list = [], start_folder: str = None
    ) -> list:
        """
        Recursively builds a folder tree.

        Note: This could be a contraction algo, but why bother?
        """
        if start_folder:
            pass
        elif self.folders:
            start_folder = self.drive_str
        else:
            return []

        level_folders = []  # all the folders within this level
        for each_folder in self.folders:
            if start_folder == each_folder.parents:
                each_folder.path = path + [each_folder.name]
                level_folders.append(
                    (
                        each_folder,
                        self.construct_folder_tree(
                            start_folder=each_folder.id, path=each_folder.path
                        ),
                    )
                )

        return level_folders

    def construct_navbar(self) -> List[Tuple[str, str]]:
        for doc in self.drive.docs():
            if doc.name == "Navbar":
                navbar = doc
                break
        else:
            print("Navbar not found!")  # TODO Handle this better
            return []

        gsimplify.docs.Docs(navbar, self.creds)

        sections = navbar.content['body']['content']

        navbar_items = []

        for section in sections:
            if 'paragraph' not in section:
                continue
            for element in section['paragraph']['elements']:
                if 'textRun' not in element:
                    continue
                if 'link' not in element['textRun']['textStyle']:
                    continue
                name = element['textRun']['content']
                link = element['textRun']['textStyle']['link']['url']
                navbar_items.append((name, self.adjust_relative_link(link)))

        return navbar_items

    def find_folder(
        self, folder: str
    ) -> Union[gsimplify.typedefs.DocType, gsimplify.typedefs.Drive]:
        if folder == self.drive_str:
            return gsimplify.typedefs.Drive()

        for each_folder in self.folders:
            if each_folder.id == folder:
                return each_folder

    def fetch_commit(self) -> str:
        """
        Fetches latest commit hash of repo.
        """
        repo = git.Repo(search_parent_directories=True)
        commit = repo.head.object.hexsha
        return commit[:6]

    def build(self, visual=True):
        """
        Loops and builds static website.
        """
        for each_doc in self.drive.docs(public=True):
            path = self.path_join(
                self.find_folder(each_doc.parents),
                prefix=f"./build/{self.fetch_commit()}",
            )

            if not os.path.exists("./build"):
                os.mkdir("./build")

            if not os.path.exists(path):
                os.mkdir(path)

            with open(path + each_doc.pointer.lower() + ".html", "w") as f:
                f.write(self.render(each_doc))

        for each_file in self.drive.media(public=True):
            path = self.path_join(
                self.find_folder(each_file.parents),
                prefix=f"./build/{self.fetch_commit()}",
            )
            if not os.path.exists(path):
                os.mkdir(path)

            with open(path + each_file.pointer.lower(), "wb") as f:
                file_handler = self.drive.service.files().get_media(fileId=each_file.id)
                downloader = MediaIoBaseDownload(f, file_handler)  # from google's docs

                if visual:
                    pbar = tqdm.tqdm(total=100)

                done = False
                while done is False:
                    status, done = downloader.next_chunk()

                    if visual:
                        pbar.update(int(status.progress()) * 100)

                if visual:
                    pbar.close()

        if self.assets_dir:
            if os.path.exists(f"./build/{self.fetch_commit()}/assets/"):
                shutil.rmtree(f"./build/{self.fetch_commit()}/assets/")
            shutil.copytree(self.assets_dir, f"./build/{self.fetch_commit()}/assets/")

        shutil.rmtree("./build/latest", ignore_errors=True)
        shutil.copytree(f"./build/{self.fetch_commit()}", "./build/latest")


    def render(self, doc, template="example.html"):
        document = gsimplify.docs.Docs(doc, self.creds, self.navbar_items, self)
        return document.render(self.templator.get("example.html"))



from jinja2 import Environment, FileSystemLoader, Template, select_autoescape


class Templates:
    def __init__(self, templates_dir=""):
        self.env = Environment(
            loader=FileSystemLoader(templates_dir),
            autoescape=select_autoescape(["html", "xml"]),
        )

    def get(self, template: str) -> Template:
        return self.env.get_template(template)



from typing import List, Optional, Tuple

from pydantic import BaseModel


class DriveObj(BaseModel):
    name: str
    id: str
    kind: str
    mime_type: str
    drive_id: str
    parents: str

    public: Optional[bool]
    draft: Optional[bool]
    meta: Optional[bool]
    webViewLink: Optional[str]
    pointer: Optional[str]


class DocType(DriveObj):
    title: Optional[str]
    content: Optional[dict]
    webViewLink: Optional[str]
    ast: Optional[List[Tuple[str, List]]]


class FolderType(DriveObj):
    path: Optional[List[str]]


class Drive(BaseModel):
    path: List[str] = []



from typing import Optional

from google.auth.transport.requests import Request
from google.oauth2 import service_account
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import Resource, build
from pydantic import BaseModel

from gsimplify.typedefs import DocType, FolderType, DriveObj

"""
TODO: Fix potential race conditions with a one-time fetch of drive objects
"""

"""
TODO: Fix potential race conditions with a one-time fetch of drive objects
"""


class Drive:
    def __init__(self, drive_id: str, creds):
        self.drive_id = drive_id
        self.service: Resource = build("drive", "v3", credentials=creds)
        self.drive_info = self.service.teamdrives().get(teamDriveId=drive_id).execute()
        self.files = self.contains()

    def contains(self):
        drive_objs = (
            self.service.files()
            .list(
                corpora="teamDrive",
                supportsTeamDrives=True,
                teamDriveId=self.drive_id,
                includeTeamDriveItems=True,
                fields="*",  # TODO: restrict scope of fields to cut down on build times
            )
            .execute()
            .get("files")
        )

        container = []
        for each_obj in drive_objs:
            if each_obj.get("mimeType") == "application/vnd.google-apps.folder":
                container.append(
                    FolderType(
                        kind=each_obj.get("kind"),
                        id=each_obj.get("id"),
                        name=each_obj.get("name"),
                        mime_type=each_obj.get("mimeType"),
                        drive_id=each_obj.get("driveId"),
                        parents=each_obj.get("parents")[0],
                    )
                )
            else:
                if each_obj.get("name").startswith("PUBLIC:"):
                    public = True
                    draft = False
                elif each_obj.get("name").startswith("DRAFT:"):
                    public = False
                    draft = True
                else:
                    public = False
                    draft = False

                if each_obj.get("mimeType") == "application/vnd.google-apps.document":

                    container.append(
                        DocType(
                            kind=each_obj.get("kind"),
                            id=each_obj.get("id"),
                            name=each_obj.get("name"),
                            mime_type=each_obj.get("mimeType"),
                            drive_id=each_obj.get("driveId"),
                            draft=draft,
                            public=public,
                            pointer=" ".join(each_obj.get("name").split(": ")[1:]),
                            parents=each_obj.get("parents")[0],
                            webViewLink=each_obj.get("webViewLink"),
                        )
                    )

                else:
                    container.append(
                        DriveObj(
                            kind=each_obj.get("kind"),
                            id=each_obj.get("id"),
                            name=each_obj.get("name"),
                            mime_type=each_obj.get("mimeType"),
                            drive_id=each_obj.get("driveId"),
                            pointer=" ".join(each_obj.get("name").split(": ")[1:]),
                            parents=each_obj.get("parents")[0],
                            draft=draft,
                            public=public,
                            webViewLink=each_obj.get("webViewLink"),
                        )
                    )

        return container

    def docs(self, public: Optional[bool] = None, draft: Optional[bool] = None):
        selector = self._selector(public, draft)

        return [x for x in self.files if isinstance(x, DocType) and selector(x)]

    def media(self, public: Optional[bool] = None, draft: Optional[bool] = None):
        selector = self._selector(public, draft)
        return [
            x
            for x in self.files
            if not isinstance(x, FolderType)
            and not isinstance(x, DocType)
            and selector(x)
        ]

    def folders(self):
        return [x for x in self.files if isinstance(x, FolderType)] + [
            FolderType(
                kind="",
                id=self.drive_id,
                name="",
                mime_type="",
                drive_id=self.drive_id,
                pointer="",
                parents="",
                path=[]
            )]

    def _selector(self, public: Optional[bool] = None, draft: Optional[bool] = None):
        if public:
            public_selector = lambda x: x if x.public else False
        elif public == False:
            public_selector = lambda x: x if not x.public else False
        else:
            public_selector = lambda x: x

        if draft:
            draft_selector = lambda x: public_selector(x) and public_selector(x).draft
        elif public == False:
            draft_selector = lambda x: public_selector(x) and not public_selector(x).draft
        else:
            draft_selector = lambda x: public_selector(x)


        return draft_selector



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gsimplify']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=2.11.2,<3.0.0',
 'gitpython>=3.1.11,<4.0.0',
 'google-api-python-client>=1.12.5,<2.0.0',
 'google-auth-httplib2>=0.0.4,<0.0.5',
 'google-auth-oauthlib>=0.4.2,<0.5.0',
 'pydantic>=1.7.2,<2.0.0',
 'tqdm>=4.53.0,<5.0.0']

setup_kwargs = {
    'name': 'gsimplify',
    'version': '0.0.7',
    'description': '',
    'long_description': '# gsimplify\nA Google docs to html compiler.\n',
    'author': 'echapman2022',
    'author_email': 'ethan.chapman0@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ChoateProgrammingUnion/simplify',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)



import collections
import pprint
from typing import List, Optional, Tuple

from google.auth.transport.requests import Request
from google.oauth2 import service_account
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import Resource, build
from jinja2 import Template
from pydantic import BaseModel

from gsimplify.build import Builder
from gsimplify.typedefs import DocType, FolderType, Drive

DEPTH_LEVELS = ["TITLE", "HEADING_1", "HEADING_2", "HEADING_3", "NORMAL_TEXT"]
HTML_DEPTH_LEVELS = ["h1", "h2", "h3", "h4", "p"]


class Docs:
    def __init__(self, doc: DocType, creds, navbar_items: List[Tuple[str, str]] = None, builder: Builder = None):
        self.document = doc
        self.service = build("docs", "v1", credentials=creds)
        self.get(doc.id)

        self.navbar_items = navbar_items
        if self.navbar_items is None:
            self.navbar_items = []

        self.builder = builder

        self.document.ast = self.parse()

    def __str__(self):
        return pprint.pformat(self.document)

    def __repr__(self):
        return str(self)

    def render(self, template: Template) -> str:
        html_tree = self.ast_to_html_tree(self.document.ast)
        return template.render(**self.document.dict(), html_tree=html_tree, navbar_items=self.navbar_items)

    def get(self, id=str) -> dict:
        """
        Grabs the document from Google Docs
        """
        self.document.content = self.service.documents().get(documentId=id).execute()
        return self.document.content

    def parse(self):
        """
        Parses document into a dict-like parse tree.
        """
        ast = self.test_parse(self.document.content.get("body").get("content"))
        self.document.title = ast[0][0]

        return ast

    def recursive_parse(self, structure: list, depth=0) -> dict:
        HTML_DEPTH_LEVELS = [""]

        """
        Recursive/monadic parser.
        """
        level = DEPTH_LEVELS[depth]
        ast = collections.OrderedDict()
        pretext = []
        posttext = []
        count = 0
        for outer_count, section in enumerate(structure):
            stop_loop = True
            if "paragraph" in section:
                paragraph = section.get("paragraph")
                style = paragraph.get("paragraphStyle").get("namedStyleType")
                if style == level:
                    content = "".join(
                        [
                            elm.get("textRun").get("content")
                            for elm in paragraph.get("elements")
                        ]
                    )

                    stop_count = 0
                    for inner_count, lookahead_section in enumerate(
                        structure[count + 1 :]
                    ):
                        if "paragraph" in lookahead_section:
                            inner_style = (
                                lookahead_section.get("paragraph")
                                .get("paragraphStyle")
                                .get("namedStyleType")
                            )
                            if (
                                inner_style == style and stop_count == 0
                            ):  # first encounter of equal depth
                                stop_count = inner_count
                                stop_loop = False
                                break
                    else:
                        stop_count = len(structure)

                    if count:
                        pretext = self.recursive_parse(structure[:count], depth=depth)
                    else:
                        pretext = collections.OrderedDict()

                    if count + 1 - stop_count:
                        posttext = self.recursive_parse(
                            structure[count + 1 : stop_count], depth=depth
                        )
                    else:
                        posttext = collections.OrderedDict()

                    ast[content] = (pretext, posttext)

                    if stop_loop:  # done
                        break
                    else:  # next elements are equal depth
                        del structure[:stop_count]
                        count = 0
                        continue  # skip the count incrementer
            count += 1
        else:
            if depth + 1 < len(DEPTH_LEVELS):
                ast[""] = (
                    collections.OrderedDict(),
                    self.recursive_parse(structure, depth=depth + 1),
                )
        return ast

    def test_parse(self, structure: list) -> List[Tuple[str, List]]:
        ast = []

        for section in structure:
            if "paragraph" not in section:
                continue

            paragraph = section.get("paragraph")
            style: str = paragraph.get("paragraphStyle").get("namedStyleType")

            def text_run_to_str(text_run: dict) -> str:
                if "link" in text_run["textStyle"]:
                    link = text_run["textStyle"]["link"]["url"]
                    if self.builder is not None:
                        link = self.builder.adjust_relative_link(link)
                    return f'<a href="{link}">{text_run["content"]}</a>'
                return text_run['content']

            content = "".join(
                [text_run_to_str(elm.get("textRun")) for elm in paragraph.get("elements")]
            ).rstrip()

            if len(content) < 1:
                continue

            depth_level = DEPTH_LEVELS.index(style)

            relative_depth_level = depth_level
            ast_slice = ast
            while relative_depth_level > 0:
                relative_depth_level -= 1
                if len(ast_slice) == 0:
                    ast_slice.append((None, []))
                ast_slice = ast_slice[-1][1]

            ast_slice.append((content, []))

        assert len(ast) == 1

        return ast

    def ast_to_html_tree(self, ast: List[Tuple[str, List]], depth=0):
        sectioned_levels = ["h2"]
        section_start = '<section>'
        section_end = '</section>'

        result = []

        for section in ast:
            if section[0] is None:
                result += self.ast_to_html_tree(section[1], depth+1)
                continue

            text_type = HTML_DEPTH_LEVELS[depth]
            sectioned = text_type in sectioned_levels

            if sectioned:
                result.append(section_start)

            result.append(f"<{text_type}>{section[0]}</{text_type}>")

            if text_type == "h1":
                result.append('<div class="main">')

            result += self.ast_to_html_tree(section[1], depth+1)

            if text_type == "h1":
                result.append('</div>')

            if sectioned:
                result.append(section_end)

        return result
